#!/bin/bash

shellfu import pretty


#
# Path where to look for files
#
# If filename does not contain slash, it is looked up in each
# (or all, based on strategy--see inigrep doc) path in this
# list.  The list is colon-separated and non-dirs as well as
# empty strings are silently ignored.
#
SHELLFU_INIGREP_PATH="${SHELLFU_INIGREP_PATH:-}"

#
# Expected filename extension (for guessing from -p path head)
#
# If no filename to read is given, inigrep will guess filename
# as the path head plus this suffix (e.g. `foo.ini` for
# `inigrep -p foo.bar.baz`)
#
SHELLFU_INIGREP_SUFFIX="${SHELLFU_INIGREP_SUFFIX:-.ini}"


_inigrep__fltcmt() {
    #
    # Just strip comments
    #
    perl -ne '
        chomp;
        next if m/^[[:space:]]*[#;]/;
        s/  *#.*//;
        s/  *;.*//;
        print "$_\n" if $_;
    '
}


_inigrep__fltkey() {
    #
    # Filter values from `key = value pairs`
    #
    # In listing mode, prints *key names* instead
    #
    while IFS= read -r line;                        # '  key  =  val  '
    do
        line="${line##*([[:space:]])}"              # line='key  =  val  '
        key="${line%%*([[:space:]])=*}"             # key='key'
        value="${line##$key*([[:space:]])=}"        # value='  val  '
        $strict || value="${value##*([[:space:]])}" # value='val'
        test "$key" = "$wntkey" || continue
        $list || echos "$value"
        $list && echos "$line"
    done
}


_inigrep__fltsct() {
    #
    # Filter out only the wanted section body
    #
    # In listing mode prints also section name
    #
    local sct_ok=false
    local sct_name
    local line
    while IFS= read -r line;
    do
        if [[ $line =~ ^[[:space:]]*\[[^]]*\].* ]];     # IOW: 'spaces[name]anything'
        then
            # hack out only the name between '[' and ']'
            sct_name=$line
            sct_name=${sct_name##*([[:space:]])[}   # drop leading space
            sct_name=${sct_name%%]*}                # and trailing anything
            if test "$sct_name" == "$wntsct";
            then
                sct_ok=true;                    # our section
                $list && echos "[$sct_name]";   # inform lister
            else
                sct_ok=false
            fi
            continue
        else
            $sct_ok || continue
            $strict || line="${line##*([[:space:]])}"
            echos "$line"
        fi
    done
}


_inigrep__fltlst() {
    #
    # Filter only section/keys/path names from the stream
    #
    local key="__NOKEY__" line="" sct="__NOSCT__"
    while read -r line;
    do
        case "$line" in
            \[*\]) sct=${line#[}; sct=${sct%]}; key=__NOKEY__     ;;
            *=*)   key="${line%%*([[:space:]])=*}" ;;   # key='key'
            *)     continue ;;                          # whatever
        esac
        case "$listobj:$sct:$key" in
            p*:__NOSCT__:*)         echos "$sct.$key" ;;
            p*:*:__NOKEY__)         continue          ;;
            p*:*:*)                 echos "$sct.$key" ;;
            s*:__NOSCT__:*)         echos "$sct"      ;;
            s*:*:__NOKEY__)         echos "$sct"      ;;
            s*:*:*)                 continue          ;;
            k*:__NOSCT__:__NOKEY__) continue          ;;
            k*:__NOSCT__:*)         echos "$key"      ;;
            k*:*:__NOKEY__)         continue          ;;
            k*:*:*)                 echos "$key"      ;;
        esac
    done
}


_inigrep__merge() {
    #
    # Take paths and applying merge strategy, load file(s)
    #
    local path
    debug -v strategy
    while read -r path;
    do
        test -f "$path" || continue
        case $strategy in
            first)
                debug "winner: $path"
                cat "$path"
                cat >/dev/null  # throw away rest of paths
                ;;
            join)
                echo "# file: ${path/$HOME/~}"
                cat "$path" 2>/dev/null
                ;;
        esac
    done
}


_inigrep__load() {
    #
    # Find file(s) candidates and load them (or use *__merge)
    #
    test -n "$1" || _inigrep__load "${wntsct%%.*}$SHELLFU_INIGREP_SUFFIX"
    local arg trydir trypath
    for arg in "$@";
    do
        case $arg in
            -|*/*)      # stdin, or path (with slash)
                cat "$arg"
            ;;
        *)              # name given, find all its incarnations
            debug -v SHELLFU_INIGREP_PATH
            echos "$SHELLFU_INIGREP_PATH" \
              | tr ':' '\n' \
              | while read -r trydir;
                do
                    test -n "$trydir" || continue
                    trypath="$trydir/$arg"
                    echos "$trypath"
                done \
              |  _inigrep__merge
            ;;
        esac
    done
    true
}


inigrep() {
    #
    # A flexible (not just) .ini file reader
    #
    # Usage:
    #   inigrep [-S] [-1] [-j] -p section.key [file...]
    #   inigrep [-S] [-1] [-j] [-s section] [-k key] [file...]
    #
    #
    # Selection
    # =========
    #
    # Probably most apparent use case is selection of key from
    # a section of INI file--the first usage form.  Note that
    # *section* itself may contain dots; only last part after all
    # dots is parsed as *key*.
    #
    # The second form lets you provide only *section* to select
    # only that section (leaving any key=value pairs intact),
    # or provide only *key* to select only that key from *all*
    # sections in file (if there are any).  This in fact extends
    # use of this function beyond just INI-like format: if you
    # are brave enough, you could e.g. query a shell script for
    # a variable assignment (i.e. key=value pair).
    #
    # If suitable key is found at multiple lines, all values
    # are printed, which allows for creating multi-line values.
    # Providing *-1* argument, however, always prints only one
    # line.
    #
    # Leading/trailing spaces and empty lines are removed from
    # output by default, but strict mode (*-S* or *--strict*
    # switch) changes this so that values after `=` are left
    # intact.  Using strict mode when not selecting key or
    # section results in preserving empty lines and comments
    # as well.
    #
    #
    # File arguments
    # ==============
    #
    # If omitted, *file* argument is inferred by taking part of
    # section name before first dot and appending value of
    # `$SHELLFU_INIGREP_SUFFIX`, (".ini" by default).
    #
    # Each *file* argument is then processed as follows:
    #
    #  *  `-` (single dash) is interpreted as reading from
    #     STDIN.
    #
    #  *  If argument contains slash, it is expanded as a regular
    #     path (relative or absolute).
    #
    #  *  Otherwise, it is taken as filename and searched for
    #     in directories given in `$SHELLFU_INIGREP_PATH`.  (This can
    #     yield more than one path, which is equivalent as if
    #     all paths were provided.)
    #
    #     Not all files expanded based on `$SHELLFU_INIGREP_PATH`
    #     are read by default; reading is governed by "merge
    #     strategy": the default strategy "first" reads only
    #     the first existing file.
    #
    #     "join" strategy on the other hand, means that any
    #     files are simply concatenated and prefixed with
    #     comment containing path to the file.  (This means that
    #     if a section is queried that is present in both files,
    #     it is effectively concatenated as well.)
    #
    #     Note that the comment can be seen only in strict mode
    #     and will be included even for non-existent files.
    #
    #     The merge strategy can be specified using parameter
    #     `--merge-strategy`, while `-j` or `--join` is a
    #     shorthand for `--merge-strategy join`
    #
    #
    # Examples
    # ========
    #
    # Following calls are equivalent
    #
    #     inigrep -p foo.bar.baz
    #     inigrep -p foo.bar.baz foo.ini
    #
    # and both result in reading of key *baz* from section *foo.bar*
    # in file *foo.ini*, which is selected from *SHELLFU_INIGREP_PATH*.
    # Should there be more foo.ini's, the first is selected.
    # Using `-j` switch
    #
    #     inigrep -j -p foo.bar.baz
    #
    # would cause all foo.ini's on *SHELLFU_INIGREP_PATH* be
    # concatenated instead.
    #
    #     inigrep -s qux ./corge.ini
    #
    # will open corge.ini in current directory, and print every
    # range of lines delimited by line containing only `[qux]`
    # and another section delimiter or end of file.  Comments and
    # empty lines will be stripped.
    #
    #     cat graply | inigrep -S -k grault -
    #
    # will split every line in graply that starts with word
    # "grault" surrounded by any amount of whitespace and
    # followed by equals sign, printing part after the equals
    # sign.  The -S (strict mode) will preserve *right hand
    # side*, ie. will not strip any spaces after "=".
    #
    local wntsct=""             # desired section name
    local wntkey=""             # desired key name
    local list=false            # listing mode?
    local listobj="paths"       # what to list (paths/sections/keys)
    local strict=false          # strict mode?
    local sct_ok=true           # are we reading a desired section
    local one_line=false        # limit to single line
    local strategy="first"      # merge strategy
    local fltpipe=""            # filter pipe
    while true; do case $1 in
        --)                                         break   ;;
        -1|--one-line) one_line=true;               shift 1 ;;
        -j|--join)     strategy="join";             shift 1 ;;
        -k|--key)      wntkey="$2";                 shift 2 || return 2 ;;
        -p|--path)     wntkey="${2##*.}";
                       wntsct="${2%.$wntkey}";      shift 2 || return 2 ;;
        -s|--section)  wntsct="$2"; sct_ok=false;   shift 2 || return 2 ;;
        -S|--strict)   strict=true;                 shift 1 ;;
        -l)            list=true; listobj=paths;    shift 1 ;;
        -L|--list)     list=true; listobj=$2;       shift 2 || return 2 ;;
        --merge-strategy) strategy="$2";            shift 2 || return 2 ;;
        *)                                                      break   ;;
    esac done
    case "$strategy" in
        first|join)                                   : ;;
        *) warn "invalid merge strategy: $strategy"; return 2 ;;
    esac
    case $listobj in
        p|paths)    : ;;
        k|keys)     : ;;
        s|sections) : ;;
        *) warn "invalid object type to list: $listobj"; return 2 ;;
    esac
    debug -v wntkey wntsct list listobj
    debug "\$*='$*'"
    $strict           || fltpipe+=" | _inigrep__fltcmt"
    test -n "$wntsct" && fltpipe+=" | _inigrep__fltsct"
    test -n "$wntkey" && fltpipe+=" | _inigrep__fltkey"
    $list             && fltpipe+=" | _inigrep__fltlst"
    $one_line         && fltpipe+=" | head -1"
    $strict           || fltpipe+=" | grep ."
    test -z "$fltpipe" && fltpipe="cat"
    fltpipe="${fltpipe# | }"
    debug -v fltpipe
    local restore_extglob=$(shopt -p extglob)
    shopt -s extglob
    _inigrep__load "$@" | eval "$fltpipe"
    $restore_extglob
}

_inigrep__nl2colon() {
    #
    # Convert newline-based list of paths to colon:based:list
    #
    # Empty paths must not be included in the resulting list,
    # so we need to drop them and also get the colons right.
    #
    local idx=0     # current item index (zero-based)
    local path
    while read -r path;
    do
        test -z "$path" && continue
        test $idx -gt 0 && echo -n ':'
        echo -n "$path"
        ((idx++))
    done
}

inigrep__mkpath() {
    #
    # Helper to assemble SHELLFU_INIGREP_PATH
    #
    # Usage:
    #
    #     inigrep__mkpath location...
    #
    # For each location, print colon-delimited list of
    # directories.  If location ends with "/ini.d", list of
    # subfolders, sorted by C locale is printed--this allows
    # for "modular" configuration.  Otherwise the location
    # is printed.  Non-directory locations are left out.
    #
    local location      # one location argument
    local path          # one path listed
    for location in "$@";
    do
        test -d "$location" || continue
        case "$location" in
            */ini.d)    # modular location--sort subfolders
                find -L "$location" -mindepth 1 -maxdepth 1 -type d \
                  | LC_ALL=C sort
                ;;
            *)
                echo "$location"
                ;;
        esac
    done \
      | _inigrep__nl2colon
}
