#!/bin/bash

shellfu import pretty

__valid_pos_int() {
    local val="$1"
    local def="$2"
    test -z "$val"      && { echo "$default"; return 1; }
    test "$val" -gt 0   && { echo "$val";     return 0; }
    warn "not a positive int, using def: '$def' instead of '$val'"
    echo "$default"
    return 1
}

__valid_int_rule() {
    local rul="$1"
    local def="$2"
    test -z "$val"           && { echo "$default"; return 1; }
    bash -n <<<"test 0 $rul" && { echo "$rul";     return 0; }
    warn "not a valid int rule, using default: '$def' instead of '$rul'"
    echo "$default"
    return 1
}


wait_until() {
    #
    # Wait until command succeeds (or fails)
    #
    # Repeat calling a command until it succeeds or exits with
    # acceptable status.  For example:
    #
    #     wait_until -t 90 my_command
    #     wait_until -t 90 -d 30 my_expensive_command
    #     wait_until -x 7 command_that_should_exit_with_7
    #     wait_until -X "-lt 7" command_that_should_exit_with_less_than_7
    #
    # Note that polling happens synchronously so if your
    # command blocks for more than specified delay, frequency
    # and timeout cannot be guaranteed.  If you need to assure
    # more predictable behavior, limit your command with
    # timeout(1)
    #
    local delay             # delay between polls in s
    local rule              # rule to check on exit status (`-eq 0`, `-lt 7`...)
    local timeout           # timeout to give up in s
    local usr_delay usr_rule usr_timeout    # user given values
    local es_test           # to hold final exit status check
    local def_delay=1       # defaults... vv
    local def_rule='-eq 0'
    local def_timeout=10
    while true; do case "$1" in
        -x) usr_rule="-eq $2";  shift; shift ;;
        -X) usr_rule="$2";      shift; shift ;;
        -d) usr_delay="$2";     shift; shift ;;
        -t) usr_timeout="$2";   shift; shift ;;
        --)                     shift; break ;;
        *)                             break ;;
    esac done
    delay="$(__valid_pos_int "$usr_delay" "$def_delay")"
    rule="$(__valid_int_rule "$usr_rule" "$def_rule")"
    timeout="$(__valid_pos_int "$usr_timeout" "$def_timeout")"
    think "waiting for: $*"
    local elapsed=0
    local start=$(date +%s)
    while true;
    do
        sleep $delay
        local now=$(date +%s)
        local elapsed=$((now - start))
        local es=0
        debug "\$*='$*'"
        "$@";
        es=$?
        debug -v es
        es_test="test $es $rule"
        if eval "$es_test";
        then
            return 0;
        fi
        if test $elapsed -gt $timeout;
        then
            warn "timeout reached ($elapsed>$timeout) when waiting for ($es $rule): $*"
            return 1
        fi
        think "waiting for another $delay s"
    done
}
